from __future__ import annotations

from collections import OrderedDict
from logging import getLogger
from typing import Any
from urllib.parse import quote

from requests import Response, Session as HTTPSession

from tradernet.string_utils import StringUtils


class NetUtils(StringUtils):
    __slots__ = ('session', 'timeout', 'logger')

    def __init__(self) -> None:
        self.session: HTTPSession | None = None
        self.timeout = 300
        self.logger = getLogger(self.__class__.__name__)

    def http_build_query(self, dictionary: dict[str, Any]) -> str:
        """
        Full analogue of the PHP function `http_build_query`.

        Parameters
        ----------
        dictionary : dict[str, Any]
            Any dictionary.

        Returns
        -------
        str
            The query built.
        """
        result = OrderedDict()

        for key, value in dictionary.items():
            if isinstance(value, list):
                for list_key, list_value in self.flatten_list(value).items():
                    result[f'{key}{list_key}'] = list_value

            elif isinstance(value, dict):
                for dict_key, dict_value in self.flatten_dict(value).items():
                    result[f'{key}{dict_key}'] = dict_value

            else:
                result[key] = value

        return self.str_from_dict(self.apply_to_dict(
            lambda va: quote(va) if isinstance(va, str) else va, result
        ))

    def request(
        self,
        method: str,
        url: str,
        headers: dict[str, str] | None = None,
        params: str | dict[str, Any] | None = None,
        data: bytes | str | dict[str, Any] | None = None
    ) -> Response:
        if not self.session:
            self.session = HTTPSession()

        response = self.session.request(
            method,
            url,
            headers=headers,
            params=params,
            data=data,
            timeout=self.timeout or None
        )
        self.logger.debug('Response url %s', response.url)

        try:
            response.raise_for_status()
        except Exception:
            self.logger.exception('Got request exception')
            raise

        return response



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal

from tradernet.symbols.base_option_symbol import BaseOptionSymbol
from tradernet.symbols.option_properties import OptionProperties


class TraderNetOption(BaseOptionSymbol):
    @staticmethod
    def decode_notation(symbol: str) -> OptionProperties:
        """
        +AAPL.18MAR2022.P150 -> (AAPL, '', -1, 150, 2022-03-18, 18MAR2022)
        """
        assert symbol.startswith('+'), \
            'Entered symbol is not a TraderNet option!'

        name_sections = symbol[1:].split('.')  # removing leading '+'

        for idx, section in enumerate(name_sections):
            try:
                expiration = TraderNetOption.decode_date(section)
                break
            except ValueError:
                continue
        else:
            raise RuntimeError('Entered symbol is not a TraderNet option!')

        # `idx` is the index of the expiration date
        assert 1 < len(name_sections[idx:]) < 4, \
            f'Entered symbol {symbol} is not a TraderNet option!'

        symbolic_expiration = name_sections[idx]
        # Everything before expiration is ticker
        ticker = '.'.join(name_sections[:idx])
        # The TraderNet option notation has no location
        location = ''

        if name_sections[idx + 1][0] == 'P':
            right = -1
        elif name_sections[idx + 1][0] == 'C':
            right = 1
        else:
            raise RuntimeError('Entered symbol is not a TraderNet option!')

        strike = Decimal('.'.join(name_sections[idx + 1:])[1:])

        return OptionProperties(
            ticker, location, right, strike, expiration, symbolic_expiration
        )

    @staticmethod
    def encode_date(conventional_date: str | date | datetime) -> str:
        if isinstance(conventional_date, str):
            conventional_date = date.fromisoformat(conventional_date)

        return conventional_date.strftime('%d%b%Y').upper()

    @staticmethod
    def decode_date(symbolic_date: str) -> date:
        return datetime.strptime(symbolic_date, '%d%b%Y').date()



from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from json import loads as json_loads
from typing import Any

from tradernet.ws_utils import WSUtils
from tradernet.string_utils import StringUtils
from tradernet.core import TraderNetCore


class TraderNetWSAPI(WSUtils, StringUtils):
    __slots__ = ('url',)

    def __init__(self, api: TraderNetCore) -> None:
        super().__init__()
        self.url = api.get_websocket_url()

    async def quotes(
        self,
        symbols: str | Sequence[str]
    ) -> AsyncIterator[dict[str, Any]]:
        """
        Subscribing quotes updates.

        Parameters
        ----------
        symbols : str | Sequence[str]
            A sequence of symbols or a single symbol.

        Yields
        ------
        AsyncIterator[dict[str, Any]]
            Quote updates.
        """
        if isinstance(symbols, str):
            symbols = [symbols]
        query = self.stringify(['quotes', symbols])

        async for message in self.get_stream(query, self.url):
            event, data = json_loads(message)
            if event == 'q':
                yield data

    async def market_depth(self, symbol: str) -> AsyncIterator[dict[str, Any]]:
        """
        Subscribe to market depth updates.

        Parameters
        ----------
        symbol : str
            A TraderNet symbol.

        Yields
        ------
        AsyncIterator[dict[str, Any]]
            Market depth updates.
        """
        query = self.stringify(['orderBook', [symbol]])

        async for message in self.get_stream(query, self.url):
            event, data = json_loads(message)
            if event == 'b':
                yield data

    async def portfolio(self) -> AsyncIterator[dict[str, Any]]:
        """
        Subscribe to portfolio updates.

        Yields
        -------
        AsyncIterator[dict[str, Any]]
            Portfolio updates.
        """
        query = self.stringify(['portfolio'])

        async for message in self.get_stream(query, self.url):
            event, data = json_loads(message)
            if event == 'portfolio':
                yield data

    async def orders(self) -> AsyncIterator[dict[str, Any]]:
        """
        Subscribing orders updates.

        Yields
        -------
        AsyncIterator[dict[str, Any]]
            Orders updates.
        """
        query = self.stringify(['orders'])

        async for message in self.get_stream(query, self.url):
            event, data = json_loads(message)
            if event == 'orders':
                yield data

    async def markets(self) -> dict[str, Any]:
        """
        Subscribing markets statuses.

        Yields
        -------
        AsyncIterator[dict[str, Any]]
            Markets statuses.
        """
        query = self.stringify(['markets'])

        async for message in self.get_stream(query, self.url):
            event, data = json_loads(message)
            if event == 'markets':
                return data
        return {}



from tradernet.client import TraderNetAPI
from tradernet.PublicApiClient import PublicApiClient as NtApi
from tradernet.symbols.tradernet_symbol import TraderNetSymbol
from tradernet.symbols.tradernet_option import TraderNetOption
from tradernet.symbols.das_option import DasOption
from tradernet.core import TraderNetCore
from tradernet.tradernet_wsapi import TraderNetWSAPI
from tradernet.trading import Trading


__all__ = [
    'TraderNetCore',
    'TraderNetAPI',
    'TraderNetWSAPI',
    'TraderNetSymbol',
    'TraderNetOption',
    'DasOption',
    'NtApi',
    'Trading'
]



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import datetime, date
from logging import getLogger

from numpy import (
    array, ndarray, diff, float64, datetime64, isfinite, isnan, log
)
from numpy.typing import NDArray
from pandas import DataFrame, to_datetime, melt

from tradernet.client import TraderNetAPI


def np_to_date(value: datetime64) -> date:
    dt = datetime.fromtimestamp(value.astype('O')/1e9)
    return date(dt.year, dt.month, dt.day)


class BaseMarketSymbol(metaclass=ABCMeta):
    """
    An abstract base class to get market data containing methods for their
    processing.

    Parameters
    ----------
    symbol : str
        A symbol name on a remote service.
    start : datetime
        The first date of the period market data to be acquired within.
    end : datetime
        The last date of the period.

    Attributes
    ----------
    symbol : str
        A symbol name on a remote service.
    api : API class instance, optional
        API is to be used to get market data.
    start : datetime
        The first date of the period market data to be acquired within.
    end : datetime
        The last date of the period.
    market_data : pandas.DataFrame
        A dataframe containing timeseries of candles: open, close, high, low.
    logger : Logger
        Saving info and debugging.
    """
    __slots__ = (
        'symbol',
        'api',
        'start',
        'end',
        'logger',
        'market_data'
    )

    def __init__(
        self,
        symbol: str,
        api: TraderNetAPI | None = None,
        start: datetime = datetime(1970, 1, 1),
        end: datetime = datetime.utcnow()
    ) -> None:
        self.symbol = symbol
        self.api = api

        # Dates interval
        self.start = start
        self.end = end

        self.logger = getLogger(self.__class__.__name__)

        self.market_data = DataFrame()

    def ffill(self, freq: str = '1D') -> None:
        today = date.today()
        try:
            self.market_data.loc[today]
        except KeyError:
            self.market_data = self.market_data.reindex(
                [*self.market_data.index.values, today]
            )

        self.market_data = self.market_data.reindex(
            to_datetime(self.market_data.index.values)
        )
        filled = self.market_data.resample(freq).ffill()
        filled = filled.reindex([*map(np_to_date, filled.index.values)])
        self.market_data = filled.ffill()

    def returns(self, kind: str = 'percent') -> NDArray[float64]:
        """
        Computing returns from market data.

        Parameters
        ----------
        kind : str
            A kind of returns. Allowed values are `percent` and `log` meaning
            per cent to previous close or natural logarithm of closes ratio.

        Returns
        -------
        result : array_like
            Symbol returns free of null values.
        """
        if 'close' not in self.market_data.columns:
            return array([])
        close_prices = self.market_data.close.values
        if kind == 'percent':
            data = diff(close_prices)/close_prices[:-1]
        elif kind == 'log':
            data = log(close_prices[1:]/close_prices[:-1])
        else:
            raise RuntimeError('Invalid return kind')
        # Filtering infinities out
        data = data[isfinite(data)]
        # Filtering non-numbers
        return data[~isnan(data)]

    def gaps(self, kind: str = 'percent') -> NDArray[float64]:
        """
        Computing gaps from market data.

        Parameters
        ----------
        kind : str
            A kind of returns. Allowed values are `percent` and `log` meaning
            per cent to previous close or natural logarithm of closes ratio.

        Returns
        -------
        result : array_like
            Symbol gaps free of null values.

        Notes
        -----
        Gap is a distance from the close price to the open price next business
        day.
        """
        columns = self.market_data.columns
        if 'open' not in columns or 'close' not in columns:
            return array([])
        open_prices = self.market_data.open.values
        close_prices = self.market_data.close.values
        if kind == 'percent':
            data = open_prices[1:]/close_prices[:-1] - 1
        elif kind == 'log':
            data = log(open_prices[1:]/close_prices[:-1])
        else:
            raise RuntimeError('Invalid return kind')
        # Filtering infinities out
        data = data[isfinite(data)]
        # Filtering non-numbers
        return data[~isnan(data)]

    def last_price(self) -> float | None:
        """
        Extracts the last non-null price from market data, either at the market
        open, or at close.

        Returns
        -------
        float, optional
            The last price if there is one.
        """
        if self.market_data.empty:
            self.logger.warning('No last price: market data is empty')
            return None

        # Melting all prices, one after another
        melted_table = melt(
            self.market_data[['open', 'close']].reset_index(),
            id_vars=['date'], var_name='moment', value_name='price'
        )
        # Creating a united dateless sequence of prices
        price_sequence = melted_table.set_index('date').sort_values(
            'moment', ascending=False   # opening prices first, then closing
        ).sort_index()['price'].values  # sorting dates, extracting prices

        if not isinstance(price_sequence, ndarray):
            # Something went wrong, the returned sequence is not a numpy array
            self.logger.warning('Failed to melt data')
            return None

        # Throwing away nulls
        existing_prices = price_sequence[~isnan(price_sequence)]
        if existing_prices.size == 0:
            self.logger.warning('All prices found are null')
            return None

        price = existing_prices[-1]  # Behold, the last price!
        self.logger.debug('Last price of %s is %s', self.symbol, price)
        return price

    @abstractmethod
    def get_data(self) -> None:
        """
        Abstract method acquiring data and assigning
        BaseMarketSymbol.market_data to them.
        """



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import ClassVar
from string import ascii_uppercase, digits

from tradernet.symbols.base_option_symbol import BaseOptionSymbol
from tradernet.symbols.option_properties import OptionProperties


class DasOption(BaseOptionSymbol):
    NUMCHAR: ClassVar[str] = f'{digits}{ascii_uppercase}'

    @staticmethod
    def decimal_to_base(number: int, base: int) -> str:
        if number < base:
            return DasOption.NUMCHAR[number]
        return DasOption.decimal_to_base(number // base, base) \
            + DasOption.NUMCHAR[number % base]

    @staticmethod
    def starting_year() -> int:
        return (date.today().year - 10) // 20 * 20 + 10

    @staticmethod
    def decode_date(symbolic_date: str) -> date:
        assert len(symbolic_date) == 3, 'Invalid date format'

        year = DasOption.starting_year() + int(symbolic_date[0], 20)
        month = int(symbolic_date[1], 16)
        day = int(symbolic_date[2], 32)

        return date(year, month, day)

    @staticmethod
    def decode_notation(symbol: str) -> OptionProperties:
        """
        +AAPL*C3I150.US -> (AAPL, US, -1, 150, 2022-03-18, C3I)
        """
        put_index = symbol.find('*')
        call_index = symbol.find('^')

        assert symbol[0] == '+' and (put_index != -1 or call_index != -1), \
            'Entered symbol is not a DAS option!'
        # Index-separator
        if call_index != -1:
            idx = call_index
            right = 1
        else:
            idx = put_index
            right = -1
        # Index following ticker
        ticker = symbol[1:idx]
        symbolic_expiration = symbol[idx + 1:idx + 4]
        expiration = DasOption.decode_date(symbolic_expiration)

        tail = symbol[idx + 4:].split('.')
        try:
            _ = int(tail[1])
            strike = Decimal('.'.join(tail[0:2]))
            location: str | None = '.'.join(tail[2:])
        except (IndexError, ValueError):
            strike = Decimal(tail[0])
            location = '.'.join(tail[1:])

        if not location:
            location = None
        return OptionProperties(
            ticker, location, right, strike, expiration, symbolic_expiration
        )

    @staticmethod
    def encode_date(conventional_date: str | date | datetime) -> str:
        if isinstance(conventional_date, str):
            conventional_date = date.fromisoformat(conventional_date)

        year = DasOption.decimal_to_base(
            conventional_date.year - DasOption.starting_year(), 20
        )
        month = DasOption.decimal_to_base(
            conventional_date.month, 16
        )
        day = DasOption.decimal_to_base(
            conventional_date.day, 32
        )

        return f'{year}{month}{day}'



from __future__ import annotations

from typing import Any, ClassVar

from tradernet.core import TraderNetCore


class PublicApiClient(TraderNetCore):
    """
    Legacy TraderNet API.

    Parameters
    ----------
    apiKey : str | None
        TraderNet public key.
    apiSecret : str | None
        TraderNet private key.
    version : int, optional
        TraderNet API version.
    """
    V1: ClassVar[int] = 1
    V2: ClassVar[int] = 2

    __slots__ = ('__version',)

    def __init__(
        self,
        apiKey: str | None = None,
        apiSecret: str | None = None,
        version: int = 1
    ) -> None:
        super().__init__(apiKey, apiSecret)
        self.__version = version

    def sendRequest(
        self,
        method: str,
        aParams: dict[str, Any] | None = None,
        format: str = 'JSON'
    ) -> Any:
        """
        Legacy interface to send a request to TraderNet.

        Parameters
        ----------
        method : str
            HTTP method.
        aParams : dict[str, Any] | None, optional
            HTTP parameters.
        format : str, optional
            Some format.

        Returns
        -------
        response : Any
            Response JSON.
        """
        return self.authorized_request(method, aParams, self.__version)
